// Author: A. Seznec March-April 2011


#include <stdio.h>
#include <cassert>
#include <string.h>
#include <inttypes.h>

using namespace std;
#include <stdlib.h>
#include "cbp3_def.h"
#include "cbp3_framework.h"

#define INCLUDEPRED
#include "predictor.h"


my_predictor *mypred;

void
PredictorInit ()
{
  
  mypred = new my_predictor ();
  assert (mypred);
}

void
PredictorReset ()
{
  // this function is called before EVERY run
  // it is used to reset predictors and change configurations


}

void
PredictorRunACycle ()
{
  // get info about what uops are processed at each pipeline stage
  const cbp3_cycle_activity_t *cycle_info = get_cycle_info ();

  // make prediction at fetch stage
  for (int i = 0; i < cycle_info->num_fetch; i++)
    {
      uint32_t fe_ptr = cycle_info->fetch_q[i];
      const cbp3_uop_dynamic_t *uop = &fetch_entry (fe_ptr)->uop;
      if (uop->type & IS_BR_INDIRECT)

	{
	  // get prediction

            
             
	  uint32_t gpred = mypred->predict_brindirect (uop->pc , uop->type );

	  // report prediction:
	  // you need to provide direction predictions for conditional branches,
	  // targets of conditional branches are available at fetch stage.
	  // for indirect branches, you need to provide target predictions.
	  // you can report multiple predictions for the same branch
	  // the framework will use the last reported prediction to calculate 
	  // misprediction penalty
	  assert (report_pred (fe_ptr, false, gpred));	
}
      // update fetch branch history
      if (uop->type & 31)

      {mypred->FetchHistoryUpdate (uop->pc , uop->type , uop->br_taken,
				    uop->br_target  );
     
           
           
      }
//#define IMMEDIATEUPDATE
#ifdef IMMEDIATEUPDATE
      if (uop->type & 31)
	{
             mypred->update_brindirect (uop->pc , uop->type , uop->br_taken, 				 uop->br_target );
              }

#endif 

}

  for (int i = 0; i < cycle_info->num_retire; i++)
    {
uint32_t rob_ptr = cycle_info->retire_q[i];
const cbp3_uop_dynamic_t *uop = &rob_entry (rob_ptr)->uop;
#ifndef  IMMEDIATEUPDATE
  if (uop->type & 31)
	{
              mypred->update_brindirect (uop->pc , uop->type , uop->br_taken, 				 uop->br_target);
              }
#endif

    }
}

void
PredictorRunEnd ()
{
 
}

void
PredictorExit ()
{
 
     
  delete mypred;

}
